import { mapSerializer, base58, SdkError } from '@metaplex-foundation/umi-core';
import { toWeb3JsMessageFromInput, fromWeb3JsMessage } from '@metaplex-foundation/umi-web3js-adapters';
import { VersionedTransaction } from '@solana/web3.js';

/* eslint-disable no-bitwise */
const TRANSACTION_VERSION_FLAG = 0x80;
const TRANSACTION_VERSION_MASK = 0x7f;
class Web3JsTransactionFactory {
  constructor(context) {
    this.context = context;
  }
  create(input) {
    const web3JsMessage = toWeb3JsMessageFromInput(input);
    const message = fromWeb3JsMessage(web3JsMessage);
    const web3JsTransaction = new VersionedTransaction(web3JsMessage, input.signatures);
    return {
      message,
      serializedMessage: this.serializeMessage(message),
      signatures: web3JsTransaction.signatures
    };
  }
  serialize(transaction) {
    return this.getTransactionSerializer().serialize(transaction);
  }
  deserialize(serializedTransaction) {
    return this.getTransactionSerializer().deserialize(serializedTransaction)[0];
  }
  serializeMessage(message) {
    return this.getTransactionMessageSerializer().serialize(message);
  }
  deserializeMessage(serializedMessage) {
    return this.getTransactionMessageSerializer().deserialize(serializedMessage)[0];
  }
  getTransactionSerializer() {
    const s = this.context.serializer;
    return {
      ...mapSerializer(s.struct([['signatures', s.array(s.bytes({
        size: 64
      }), {
        size: shortU16()
      })], ['serializedMessage', s.bytes()]]), value => value, value => ({
        ...value,
        message: this.deserializeMessage(value.serializedMessage)
      })),
      description: 'Transaction'
    };
  }
  getTransactionMessageSerializer() {
    return {
      description: 'TransactionMessage',
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const serializer = this.getTransactionMessageSerializerForVersion(value.version);
        return serializer.serialize(value);
      },
      deserialize: (bytes, offset = 0) => {
        const [version] = this.getTransactionVersionSerializer().deserialize(bytes, offset);
        const serializer = this.getTransactionMessageSerializerForVersion(version);
        return serializer.deserialize(bytes, offset);
      }
    };
  }
  getTransactionMessageSerializerForVersion(version) {
    const s = this.context.serializer;
    return s.struct([['version', this.getTransactionVersionSerializer()], ['header', this.getTransactionMessageHeaderSerializer()], ['accounts', s.array(s.publicKey(), {
      size: shortU16()
    })], ['blockhash', s.string({
      encoding: base58,
      size: 32
    })], ['instructions', s.array(this.getCompiledInstructionSerializer(), {
      size: shortU16()
    })], ['addressLookupTables', s.array(this.getCompiledAddressLookupTableSerializer(), {
      size: version === 'legacy' ? 0 : shortU16()
    })]]);
  }
  getTransactionVersionSerializer() {
    return {
      description: 'TransactionVersion',
      fixedSize: null,
      maxSize: 1,
      serialize: value => {
        if (value === 'legacy') return new Uint8Array([]);
        return new Uint8Array([TRANSACTION_VERSION_FLAG | value]);
      },
      deserialize: (bytes, offset = 0) => {
        const slice = bytes.slice(offset);
        if (slice.length === 0 || (slice[0] & TRANSACTION_VERSION_FLAG) === 0) {
          return ['legacy', offset];
        }
        const version = slice[0] & TRANSACTION_VERSION_MASK;
        if (version > 0) {
          throw new SdkError(`Unsupported transaction version: ${version}.`);
        }
        return [version, offset + 1];
      }
    };
  }
  getTransactionMessageHeaderSerializer() {
    const s = this.context.serializer;
    return s.struct([['numRequiredSignatures', s.u8()], ['numReadonlySignedAccounts', s.u8()], ['numReadonlyUnsignedAccounts', s.u8()]]);
  }
  getCompiledInstructionSerializer() {
    const s = this.context.serializer;
    return s.struct([['programIndex', s.u8()], ['accountIndexes', s.array(s.u8(), {
      size: shortU16()
    })], ['data', s.bytes({
      size: shortU16()
    })]]);
  }
  getCompiledAddressLookupTableSerializer() {
    const s = this.context.serializer;
    return s.struct([['publicKey', s.publicKey()], ['writableIndexes', s.array(s.u8(), {
      size: shortU16()
    })], ['readonlyIndexes', s.array(s.u8(), {
      size: shortU16()
    })]]);
  }
}

/**
 * Same as u16, but serialized with 1 to 3 bytes.
 *
 * If the value is above 0x7f, the top bit is set and the remaining
 * value is stored in the next bytes. Each byte follows the same
 * pattern until the 3rd byte. The 3rd byte, if needed, uses
 * all 8 bits to store the last byte of the original value.
 */
const shortU16 = () => ({
  description: 'shortU16',
  fixedSize: null,
  maxSize: 3,
  serialize: value => {
    const bytes = [];
    let remainingValue = value;
    for (;;) {
      let elem = remainingValue & 0x7f;
      remainingValue >>= 7;
      if (remainingValue === 0) {
        bytes.push(elem);
        break;
      } else {
        elem |= 0x80;
        bytes.push(elem);
      }
    }
    return new Uint8Array(bytes);
  },
  deserialize: (buffer, offset = 0) => {
    const bytes = [...buffer.slice(offset)];
    let len = 0;
    let size = 0;
    for (;;) {
      const elem = bytes.shift();
      len |= (elem & 0x7f) << size * 7;
      size += 1;
      if ((elem & 0x80) === 0) {
        break;
      }
    }
    return [len, offset + size];
  }
});

export { Web3JsTransactionFactory, shortU16 };
//# sourceMappingURL=Web3JsTransactionFactory.mjs.map
